-- abstract syntax

module Abstract where

import Prelude hiding (showList)

import Control.Monad.Writer

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set

import Text.PrettyPrint

import Util hiding (parens, brackets)
import qualified Util
import TreeShapedOrder (TSO)
import qualified TreeShapedOrder as TSO

type Name = String

data Sized = Sized | NotSized 
             deriving (Eq,Ord,Show)

data Co = Ind 
        | CoInd
          deriving (Eq,Ord,Show)

-- polarities and strict positivity ----------------------------------

data Pos 
  = SPos  -- strictly positive 
  | Pos   -- positive
  | Mixed -- unknown polarity (invariant)
  | Neg   -- negative, used internally for contravariance of sized codata  
  | NOcc  -- non-occurring, used to determine whether data type is recursive
         deriving (Eq,Ord)

instance Show Pos where
  show SPos  = "++"
  show Pos   = "+"
  show Mixed = "o"
  show Neg   = "-"
  show NOcc  = "T"

isSPos :: Pos -> Bool
isSPos SPos = True
isSPos NOcc = True
isSPos _ = False

isPos :: Pos -> Bool
isPos Pos = True
isPos x = isSPos x

polNeg :: Pos -> Pos
polNeg SPos  = Neg
polNeg Pos   = Neg
polNeg Neg   = Pos
polNeg Mixed = Mixed
polNeg NOcc  = NOcc

-- polarity composition
polComp :: Pos -> Pos -> Pos
polComp NOcc  x   = NOcc   -- most dominant
polComp x NOcc    = NOcc
polComp Mixed x   = Mixed  -- dominant except for NOcc
polComp x Mixed   = Mixed
polComp SPos  x   = x      -- neutral
polComp x SPos    = x
polComp Pos  x    = x      -- neutral except for SPos
polComp x Pos     = x
polComp Neg Neg   = Pos    -- order 2

-- polarity conjuction (infimum)
polAnd :: Pos -> Pos -> Pos
polAnd NOcc x       = x       -- most information
polAnd x NOcc       = x
polAnd x y | x == y = x       -- same information
polAnd SPos Pos     = Pos     -- SPos is more informative than Pos
polAnd Pos SPos     = Pos
polAnd _ _          = Mixed   -- remaining cases: conflicting or no info
{-
polAnd Mixed x  = Mixed   -- least information
polAnd x Mixed  = Mixed
polAnd SPos Neg = Mixed
polAnd Neg SPos = Mixed
-}

-- decoration of Pi-types --------------------------------------------

-- 1. whether argument is erased
-- further possibilities:
-- 2. polarity
-- 3. hidden

newtype Dec = Dec { erased :: Bool }
           deriving (Eq, Ord)

defaultDec = notErased
notErased = Dec False
argErased = Dec True
resurrectDec d = d { erased = False }

instance Show Dec where
    show dec = if erased dec then "[-]" else "(-)"

-- sorts -------------------------------------------------------------

data Sort = Tm   -- sort of terms, only needed for irrelevance/extraction
          | Size
          | Set  -- TODO: predicative hierarchy
          | Type -- this is Set\omega, used internally only
            deriving (Eq, Ord, Show)

leSort :: Sort -> Sort -> Bool
leSort _ Type = True
leSort Type _ = False
leSort s s'   = s == s' 

-- identifiers -------------------------------------------------------

data IdKind = Dat     -- data/codata 
            | Con Co  -- constructor ind/coind
            | Fun     -- fun/cofun 
            | Let     -- let definition
              deriving (Eq, Ord)

data DefId = DefId { idKind :: IdKind, name :: Name } 
           deriving (Eq, Ord)

instance Show DefId where
    show = name 

type MVar = Int -- metavariables are numbered

-- expressions -------------------------------------------------------

data Expr = Sort Sort    -- Size Set
          -- size type
          | Succ Expr
          | Infty
          | Max [Expr]
          -- identifiers 
          | Meta MVar    -- meta-variable
          | Var Name     -- variables are named
          | Def DefId    -- identifiers in the signature
{-
          | Con Co Name  -- constructors
          | Def Name     -- fun/cofun ?
          | Let Name     -- definition (non-recursive)
-}
          -- dependently typed lambda calculus
          | Case Expr [Clause]
          | LLet Dec Name Expr Expr Expr --local let [x : A] = t in u
          | App Expr Expr
          | Lam Dec Name Expr
          | Pi Dec Name Expr Expr
          | Sing Expr Expr -- <t : A> singleton type
          -- for extraction: erased term
          | Erased Expr
          | Irr -- for instance the term correponding to the absurd pattern
          deriving (Eq,Ord)

data Clause = Clause { clPatters :: [Pattern], clExpr :: Maybe Expr }  
            deriving (Eq,Ord,Show)

data Pattern = VarP Name         -- x
             | ConP Co Name [Pattern] -- (c ps)
             | SuccP Pattern     -- ($ p)
             | SizeP Name Name   -- (x > y)
             | UnusableP Pattern  
{- a pattern which results from matching a coinductive type and
the corresponding size index is not in the coinductive result type of
the function.  Such a pattern is not usable for termination
checking. -}
             | DotP Expr         -- .p
             | AbsurdP           -- ()
             | ErasedP Pattern   -- pattern which got erased
--             | IrrP -- pattern which got erased
               deriving (Eq,Ord)

type Case = (Pattern,Expr)


con co n = Def $ DefId (Con co) n
fun n    = Def $ DefId Fun n
dat n    = Def $ DefId Dat n
letdef n = Def $ DefId Let n

type SpineView = (Expr, [Expr])

-- collect applications to expose head
spineView :: Expr -> SpineView
spineView = aux []
  where aux sp (App f e) = aux (e:sp) f
        aux sp e = (e, sp)

test_spineView = spineView ((Var "x" `App` Var "y") `App` Var "z")
{-
-- sort expressions
set  = Sort Set
size = Sort Size
-}

type Extr = Expr -- extracted expressions

-- external reference to recursive function (outside of the body)
mkExtName n = "_" ++ n
mkExtRef  n = letdef (mkExtName n)

-- inferable term = True/False
-- not needed for types or sizes
inferable :: Expr -> Bool
inferable Var{}   = True
inferable Sort{}  = True
inferable Infty{} = True
--inferable Con{}   = True
inferable Def{}   = True
--inferable Let{}   = True
inferable App{}   = True
-- inferable Sing{}  = True  -- not with universes
inferable _       = False

-- boilerplate to extract free variables in the usual sense
freeVars :: Expr -> Set Name
freeVars Sort{}       = Set.empty
freeVars (Succ e)     = freeVars e
freeVars Infty        = Set.empty
freeVars Meta{}       = Set.empty
freeVars (Var name)   = Set.singleton name
--freeVars Con{}        = Set.empty
freeVars Def{}        = Set.empty
--freeVars Let{}        = Set.empty
freeVars (Case e cls) = Set.unions (freeVars e : map freeVarsClause cls) 
freeVars (LLet dec x a t u) 
  = freeVars a `Set.union` (Set.delete x (freeVars u)) `Set.union` freeVars t
freeVars (App f e)   = freeVars f `Set.union` freeVars e
-- freeVars (App e es)   = Set.unions (freeVars e : map freeVars es)
freeVars (Max es)     = Set.unions (map freeVars es)
freeVars (Lam _ x e)  = Set.delete x (freeVars e)
freeVars (Pi _ x a b) = freeVars a `Set.union` Set.delete x (freeVars b)
freeVars (Sing e t)   = freeVars e `Set.union` freeVars t
freeVars (Erased e)   = freeVars e

freeVarsClause :: Clause -> Set Name
freeVarsClause (Clause ps Nothing) = Set.empty
freeVarsClause (Clause ps (Just e)) = foldl (Set.\\) (freeVars e) (map (Set.fromList . patternVars) ps)

patternVars :: Pattern -> [Name]
patternVars (VarP name)   = [name]
patternVars (SizeP x y)   = [y]
patternVars (SuccP p)     = patternVars p
patternVars (ConP _ _ ps) = concat $ map patternVars ps
patternVars (DotP _)      = []
patternVars (ErasedP p)   = patternVars p
patternVars (AbsurdP)     = []

-- get all the definitions that are refered to in expression
-- this is used e.g. to check whether a (co)fun is recursive
usedDefs :: Expr -> [Name]
usedDefs (Def id) | idKind id == Fun = [name id]
usedDefs (Def id) | idKind id == Dat = [name id]
usedDefs (Def{}) = []
usedDefs (App f e) = usedDefs f ++ usedDefs e
-- usedDefs (App e es) = foldl (++) (usedDefs e) (map usedDefs es)
usedDefs (Max es) = concat (map usedDefs es)
usedDefs (Lam _ x e)  = usedDefs e
usedDefs (Sing a b) = usedDefs a ++ usedDefs b
usedDefs (Pi _ _ a b) = usedDefs a ++ usedDefs b
usedDefs (LLet dec _ e e1 e2) = usedDefs e ++ usedDefs e1 ++ usedDefs e2
usedDefs (Succ e) = usedDefs e
usedDefs (Case e cls) = foldl (++) (usedDefs e) (map ((maybe [] usedDefs) . clExpr) cls) 
usedDefs (Erased e) = usedDefs e
usedDefs Sort{} = []
usedDefs Infty  = []
usedDefs Meta{} = []
usedDefs Var{} = []
-- usedDefs Con{} = []
-- usedDefs Let{} = []
         
rhsDefs :: [Clause] -> [Name]
rhsDefs cls = foldl (\ ns (Clause ps e) -> maybe [] usedDefs e ++ ns) [] cls

-- pretty printing expressions ---------------------------------------

[precArrL, precAppL, precAppR] = [1..3]

instance Pretty Expr where
  prettyPrec _ Irr         = text "()"
  prettyPrec _ (Sort s)    = text $ show s
  prettyPrec _ Infty       = text "#"
  prettyPrec _ (Meta i)    = text $ "?" ++ show i
  prettyPrec _ (Var n)     = text n
--  prettyPrec _ (Con _ n)   = text n
  prettyPrec _ (Def id)    = text $ name id
--  prettyPrec _ (Let n)     = text n
  prettyPrec _ (Sing e t)  = angleBrackets $ pretty e <+> colon <+> pretty t
  prettyPrec k (Succ e)    = parensIf (precAppR <= k) $ 
                              text "$" <+> prettyPrec precAppR e  
  prettyPrec k (Max es)  = parensIf (precAppR <= k) $
    foldl (\ d e -> d <+> prettyPrec precAppR e) (text "max") es
  prettyPrec k (App f e)  = parensIf (precAppR <= k) $
    prettyPrec precAppL f <+> prettyPrec precAppR e 
--   prettyPrec k (App e [])  = prettyPrec k e
--   prettyPrec k (App e es)  = parensIf (precAppR <= k) $
--     foldl (\ d e -> d <+> prettyPrec precAppR e) (prettyPrec precAppL e) es
  prettyPrec k (Case e cs) = parensIf (0 < k) $
    (text "case" <+> pretty e) $$ (vlist $ map prettyCase cs)
  prettyPrec k (Lam dec x e) = parensIf (0 < k) $ 
    (if erased dec then brackets else id) (text "\\" <+> text x <+> text "->") 
      <+> pretty e
  prettyPrec k (LLet dec n t e1 e2) = parensIf (0 < k) $
    (text "let" <+> (if erased dec then lbrack else empty) <>
       text n <+> vcat [ colon <+> pretty t 
                           <> (if erased dec then rbrack else empty) 
                       , equals <+> pretty e1 ])
    $$ (text "in" <+> pretty e2)
  prettyPrec k (Pi dec x t1 t2) | null x = parensIf (precArrL <= k) $ 
    ((if erased dec then brackets (pretty t1) else prettyPrec precArrL t1) 
      <+> text "->" <+> pretty t2)
  prettyPrec k (Pi dec x t1 t2) = parensIf (precArrL <= k) $ 
    ((if erased dec then brackets else parens) $
      text x <+> colon <+> pretty t1) <+> text "->" <+> pretty t2
  prettyPrec k (Erased e) = brackets $ pretty e

prettyCase (Clause [p] Nothing)  = pretty p 
prettyCase (Clause [p] (Just e)) = pretty p <+> text "->" <+> pretty e

vlist :: [Doc] -> Doc
vlist [] = lbrace <> rbrace
vlist ds = (vcat $ zipWith (<+>) (lbrace : repeat semi) ds) $$ rbrace


instance Pretty Pattern where
  prettyPrec k = prettyPrec k . patternToExpr

instance Show Expr where
  showsPrec k e s = render (prettyPrec k e) ++ s
  -- show = render . pretty -- showExpr

instance Show Pattern where
  show = render . pretty

{- OBSOLETE
showExpr :: Expr -> String
showExpr e = 
    case e of
      Set -> "Set"
      Size -> "Size" 
      Succ e -> "($ " ++ showExpr e ++ ")"
      Infty -> "#"
      Meta i -> "?" ++ show i
      Var n -> n
      Con _ n -> n
      Def n -> n -- "δ" ++ n
      Let n -> n
      Case e cs -> "case " ++ showExpr e ++ " { " ++ showCases cs ++ " } "
      LLet n t1 e1 e2 ->
          "(let " ++ n ++ " : " ++ showExpr t1 ++ " = " ++ showExpr e1 ++ " in " ++ showExpr e2 ++ ")"  
      App e1 el -> "(" ++ showExprs (e1:el) ++ ")"
      Lam _ x e1 -> "(\\" ++ x ++ " -> " ++ showExpr e1 ++ ")"
      Sing e t -> "<" ++ showExpr e ++ " : " ++ showExpr t ++ ">"
      Pi dec x t1 t2 -> Util.parens $ (if erased dec then Util.brackets binding
                                    else if x=="" then s1 else Util.parens binding) 
                                  ++ " -> " ++ showExpr t2
         where s1 = showExpr t1
               binding = if x == "" then s1 else x ++ " : " ++ s1
      Erased e -> Util.brackets $ showExpr e 


showExprs :: [Expr] -> String
showExprs = showList " " showExpr

showPattern :: Pattern -> String
showPattern = showExpr . patternToExpr
{-
showPattern (ConP c _ ps) = Util.parens $ foldl (\ acc p -> acc ++ " " ++ showPattern p) c ps
showPattern (SuccP p) = Util.parens $ "$ " ++ showPattern p
showPattern (DotP e)  = "." ++ showExpr e
showPattern (VarP x) = x
-}
-}               

showCase (Clause [p] Nothing) = render (prettyPrec precAppR p)
showCase (Clause [p] (Just e)) = render (prettyPrec precAppR p) ++ " -> " ++ show e
showCases = showList "; " showCase 



-- substitution ------------------------------------------------------

-- parallel substitution (CAUTION! NOT CAPTURE AVOIDING!) 
-- only needed to generate destructors
-- does not substitute into patterns of a Case
parSubst :: (Name -> Expr) -> Expr -> Expr
parSubst phi e@Sort{}       = e
parSubst phi (Succ e)       = Succ (parSubst phi e)
parSubst phi e@Infty        = e
parSubst phi e@Meta{}       = e
parSubst phi (Var x)        = phi x
-- parSubst phi e@Con{}        = e
parSubst phi e@Def{}        = e
-- parSubst phi e@Let{}        = e
parSubst phi (Case e cls)   = Case (parSubst phi e) $ 
  map (\ (Clause [p] t) -> Clause [p] (fmap (parSubst phi) t)) cls
parSubst phi (LLet dec x a b c) = LLet dec x (parSubst phi a) (parSubst phi b) (parSubst phi c)
parSubst phi (App f e)      = App (parSubst phi f) (parSubst phi e)
-- parSubst phi (App e es)     = App (parSubst phi e) (map (parSubst phi) es)
parSubst phi (Max es)       = Max (map (parSubst phi) es)
parSubst phi (Lam dec x e)  = Lam dec x (parSubst phi e)
parSubst phi (Pi dec x a b) = Pi dec x (parSubst phi a) (parSubst phi b)
parSubst phi (Sing a b)     = Sing (parSubst phi a) (parSubst phi b)
parSubst phi (Erased e)     = Erased (parSubst phi e)

{- NOT NEEDED
sgSubst :: Name -> Expr -> Expr -> Expr
sgSubst x t u = parSubst (\ y -> if x == y then t else Var y) u
-}


type Subst = Map MVar Expr

-- metavariable substitution (CAUTION! NOT CAPTURE AVOIDING!) 
-- does not substitute in patterns!
subst :: Subst -> Expr -> Expr
subst phi e@Sort{}       = e
subst phi (Succ e)       = Succ (subst phi e)
subst phi e@Infty        = e
subst phi e@(Meta i) = 
  case Map.lookup i phi of
    Just e' -> e'
    _ -> e
subst phi e@Var{}        = e
-- subst phi e@Con{}        = e
subst phi e@Def{}        = e
-- subst phi e@Let{}        = e
subst phi (Case e cls)   = Case (subst phi e) $
  map (\ (Clause [p] t) -> Clause [p] (fmap (subst phi) t)) cls
subst phi (LLet dec x a b c) = LLet dec x (subst phi a) (subst phi b) (subst phi c)
subst phi (App f e)      = App (subst phi f) (subst phi e)
-- subst phi (App e es)     = App (subst phi e) (map (subst phi) es)
subst phi (Max es)       = Max (map (subst phi) es)
subst phi (Lam dec x e)  = Lam dec x (subst phi e)
subst phi (Pi dec x a b) = Pi dec x (subst phi a) (subst phi b)
subst phi (Sing a b)     = Sing (subst phi a) (subst phi b)
subst phi (Erased e)     = Erased (subst phi e)

-- declarations --------------------------------------------------

data Declaration 
  = DataDecl Name Sized Co [Pos] Telescope Type [Constructor]  -- data/codata
  | FunDecl Co [Fun]  -- mutual fun block / mutual cofun block
  | LetDecl Bool TypeSig Expr -- bool for eval ??
    deriving (Eq,Ord,Show)

data TypeSig = TypeSig { namePart :: Name, typePart :: Type }
             deriving (Eq,Ord,Show)

type Type = Expr

type Constructor = TypeSig

type TBind = (Dec,Name,Type)  -- same as TypeSig ??

type Telescope = [TBind]

type Fun = (TypeSig,[Clause])
-- extracted declarations

type EDeclaration = Declaration
type EClause = Clause
type EPattern = Pattern
type EConstructor = Constructor
type ETypeSig = TypeSig
type EFun = Fun

-- Printing declarations ---------------------------------------------

{-
instance Show Declaration where
  show = render . pretty

instance Pretty Declaration
  pretty (DataD
-}

-- pretty print a function body
prettyFun :: Name -> [Clause] -> Doc
prettyFun f cls = vlist $ map (prettyClause f) cls

prettyClause f (Clause ps Nothing) = text f <+> hsep (map (prettyPrec precAppR) ps)
prettyClause f (Clause ps (Just e)) = text f 
  <+> hsep (map (prettyPrec precAppR) ps)
  <+> equals <+> pretty e

-- Constructor analysis ----------------------------------------------

data FieldClass = Index 
                | Field (Maybe (Type, Clause)) deriving (Eq, Show)  
  -- maybe type and def of destructor
data FieldInfo = FieldInfo 
  { fDec :: Dec
  , fName  :: Name        -- "" for anonymous fields
  , fType  :: Type
--  , fLazy  :: Bool        -- lazy (coinductive occ) or strict (everything else) -- see TCM.hs ConSig
  , fClass :: FieldClass 
  } 

instance Show FieldInfo where
  show (FieldInfo dec name t fcl) = 
    (if fcl == Index then "index " else "field ") ++
    bracketsIf (erased dec) (name ++ " : " -- ++ (if lazy then "?" else "") 
                                      ++ show t) 

data ConstructorInfo = ConstructorInfo 
  { cName   :: Name 
--  , cType   :: TVal
  , cFields :: [FieldInfo]
  , cTyCore :: Type
  , cPatFam :: (Bool, [Pattern]) 
  , cEtaExp :: Bool -- prerequisite: all destructors are defined
  } deriving Show

corePat :: ConstructorInfo -> [Pattern]
corePat = snd . cPatFam

{-  For the constructor

  vcons : (n : Nat) -> (head : A) -> (tail : Vec A n) -> Vec A (suc n)

cName   = "vcons"
-- cType   = evaluation of (A : Set) -> (n : Nat) -> ...
cFields = [("n",Nat,Index),("head",A,Field),("tail",Vec A n,Field)]
cTyCore = Vec A (suc n)
cPatFam = (True, [A, suc n])
cEtaExp = True, but may be set to False later since the constructor is recursive 

We generate the destructors
  
  head : (A : Set) -> (n : Nat) -> (x : Vec A (suc n)) -> A
  head .A .n (vcons A n _head _tail) = _head
  
  tail : (A : Set) -> (n : Nat) -> (x : Vec A (suc n)) -> Vec A n
  tail .A .n (vcons A n _head _tail) = _tail

in the implementation we use "constructed_by_head" for "x"

discriminate index arguments from fields  
  - split constructor type into telescope and core
    [(n : Nat),(head : A),(tail : Vec A n)], Vec A (suc n)
  - find free variables of core: [A,n]
  - create a list of (name,type,classification) for each constructor arg,
    where classification in {index,field}

-}

{- functionality relocated to sposConstructor 
-- cannot exclude that the type is a recursive occurrence
maybeRecursiveOccurrence :: Name -> Expr -> Bool
maybeRecursiveOccurrence dataName ty =
  case (fst (spineView ty)) of
    Def (DefId Dat n) -> n == dataName
    Def (DefId (Con _) n) -> False
    Pi{} -> False
    Sort{} -> False
    Var{} -> False
    Sing e ty -> maybeRecursiveOccurrence dataName ty
    Erased e  -> maybeRecursiveOccurrence dataName e
    Case{} -> True
    LLet{} -> True
    Def (DefId Fun _) -> True -- being cautious
    Def (DefId Let _) -> True -- being cautious
-- TODO: analyze value, not expression!
-}
    
classifyFields :: Co -> Name -> Type -> [FieldInfo] 
classifyFields co dataName ty = map (classifyField fvs) tele
  where (tele, core) = typeToTele ty
        fvs = freeVars core
        classifyField fvs (dec, name, ty) = FieldInfo
          { fDec = dec
          , fName  = name
          , fType  = ty
--          , fLazy  = co == CoInd && maybeRecursiveOccurrence dataName ty
          , fClass = if name `Set.member` fvs then Index else Field Nothing
          }

-- can only eta expand if I can generate all destructors
etaExpandable :: [FieldInfo] -> Bool
etaExpandable fields = 
  all (\ f -> fClass f == Index || fDec f == notErased && fName f /= "") fields

destructorNames :: [FieldInfo] -> [String]
destructorNames fields = 
  map fName $ filter (\ f -> fClass f /= Index && fDec f == notErased) fields

analyzeConstructor :: Co -> Name -> Telescope -> Constructor -> ConstructorInfo
analyzeConstructor co dataName pars (TypeSig constrName ty) = 
  let (_, core) = typeToTele ty
      fields = classifyFields co dataName ty
      -- generate destructors
      -- choose a name for the record to destroy
      indices = filter (\ f -> fClass f == Index) fields
      indexTele = map (\ f -> (fDec f, fName f, fType f)) indices
      indexNames = map fName indices
      -- do not generated destructors for erased arguments
      destrNames = destructorNames fields
      recName = "constructed_by_" ++ constrName
      parNames = map (\ (_,x,_) -> x) pars
      parAndIndexNames = parNames ++ indexNames
      -- substitute variable "fst" by application "fst A B p"
      phi x = if x `elem` destrNames 
                then foldl App (fun x) (map Var (parAndIndexNames ++ [recName]))
                else Var x
      prefix d =  "destructor_argument_" ++ d
      -- modifiedDestrNames = map prefix destrNames
      -- TODO: Index arguments are not always before fields
      pattern = ConP co constrName (map (DotP . Var) parNames 
         ++ map (\ fi -> (case fClass fi of 
                            Index   -> DotP . Var
                            Field{} -> VarP . prefix) 
                         (fName fi)) 
              fields)
      destrType t = teleToTypeErase (pars ++ indexTele) 
                      $ (Pi notErased recName core . parSubst phi) t
      destrBody (dn) = Clause (map VarP parAndIndexNames ++ [pattern]) (Just (Var dn))
      fields' = -- if etaExpandable fields then
                  map (\ f -> if fClass f /= Index && fDec f == notErased then
                                 f { fClass = Field $ Just 
                                        ( destrType (fType f)
                                        , destrBody (prefix (fName f)) )}
                               else f) fields
                -- else fields 
  in ConstructorInfo 
       { cName = constrName
       , cFields = fields'
       , cTyCore = core
       , cPatFam = fromAllWriter $ isPatIndFamC core 
       , cEtaExp = etaExpandable fields
       } 
{-
  where classifyField fvs (dec, name, ty) = FieldInfo
          { fDec = dec
          , fName  = name
          , fType  = ty
          , fClass = if name `Set.member` fvs then Index else Field Nothing
          }
-}

-- Pattern inductive families ----------------------------------------

-- isPatIndFam takes a list of type signatures (constructor decls.)
-- and checks whether we have a pattern inductive family
-- in this case, a list of constructors with the associated
-- type indices (translated into pattern list) is returned
-- type parameters are dropped
{-
isPatIndFam :: Int -> [Constructor] -> Maybe [(Name,[Pattern])]
isPatIndFam numPars= mapM (\ tysig -> 
                             fmap (\ ps -> (namePart tysig, drop numPars ps))
                                  (isPatIndFamC (typePart tysig)))
-}

-- isPatIndFamC checks whether an expression (the type of s constructor) 
-- is of the form 
--   Gamma -> D ps
-- and returns the list ps of patterns if it is the case
isPatIndFamC :: Expr -> Writer All [Pattern]
isPatIndFamC (Def id) = return []
isPatIndFamC (App f e) = do
  ps <- isPatIndFamC f
  p  <- exprToDotPat' e
  return $ ps ++ [p]
-- isPatIndFamC (App e es) = do
--   ps  <- isPatIndFamC e
--   ps' <- mapM exprToDotPat' es
--   return $ ps ++ ps'
isPatIndFamC (Pi _ _ _ e) = isPatIndFamC e
isPatIndFamC _ = tell (All False) >> return []

-- Pattern auxiliary functions ---------------------------------------

-- extract all subpatterns of the form y > x and arrange them in a
-- TreeShapedOrder
tsoFromPatterns :: [Pattern] -> TSO Name
tsoFromPatterns ps = TSO.fromList $ concat $ map loop ps where
  loop (SizeP father son) = [(son,father)]
  loop (ConP _ _ ps)      = concat $ map loop ps
  loop (SuccP   p)        = loop p
  loop (ErasedP p)        = loop p
  loop VarP{}             = []
  loop DotP{}             = []
  loop UnusableP{}        = []
  
-- for non-dot patterns, patterns overlap if one matches against the other
-- infinity size is represented as (DotP Infty)
-- I reprogram it here, since it does not need a monad
overlap :: Pattern -> Pattern -> Bool
overlap (VarP _) p' = True
overlap p (VarP _)  = True
overlap (ConP _ c ps) (ConP _ c' ps') = c == c' && overlaps ps ps' -- only source of non-overlap
-- size patterns always overlap
overlap (SuccP p) _ = True
overlap _ (SuccP p) = True
overlap SizeP{} _   = True
overlap _ SizeP{}   = True
-- dot patterns always overlap (safe approximation)
overlap (DotP _) _ = True
overlap _ (DotP _) = True
{-
overlap (SuccP p) (SuccP p') = overlap p p'
overlap (SuccP p) (DotP Infty) = overlap p (DotP Infty)
overlap (DotP Infty) (SuccP p') = overlap (DotP Infty) p'
overlap (DotP Infty) (DotP Infty) = True
-}

overlaps :: [Pattern] -> [Pattern] -> Bool
overlaps ps ps' = and $ zipWith overlap ps ps'


-- exprToPattern is used to check whether an ind.fam. is a pattern ind.fam.
exprToPattern :: Expr -> Maybe Pattern
exprToPattern (Var n)    = return $ VarP n
exprToPattern (Def (DefId (Con co) n)) = return $ ConP co n []
-- exprToPattern (Infty)    = return $ DotP Infty -- leads to non-term in compareExpr
exprToPattern (Succ e)   = exprToPattern e >>= return . SuccP
exprToPattern (App f e) = do
  (ConP co n ps) <- exprToPattern f
  pe <- exprToPattern e
  return $ (ConP co n (ps ++ [pe]))
-- exprToPattern (App e es) = do
--   p <- exprToPattern e
--   ps <- mapM exprToPattern es
--   p `patApps` ps
exprToPattern _ = fail "exprToPattern"

{- RETIRED
patApps :: (Monad m) => Pattern -> [Pattern] -> m Pattern
patApps p [] = return p
patApps (ConP co n ps) ps' = return $ ConP co n (ps ++ ps')
patApps _ _ = fail "patApps"
-}

-- turn an expression into a pattern
-- the Bool is True if the pattern is proper, i.e., does not contain
-- DotP except (DotP Infty)
exprToDotPat :: Expr -> (Bool, Pattern)
exprToDotPat = fromAllWriter . exprToDotPat'

exprToDotPat' :: Expr -> Writer All Pattern
exprToDotPat' (Var n)    = return $ VarP n
exprToDotPat' (Def (DefId (Con co) n)) = return $ ConP co n []
exprToDotPat' (Infty)    = return $ DotP Infty
exprToDotPat' (Succ e)   = exprToDotPat' e >>= return . SuccP
exprToDotPat' e@(App f e') = do
  pf <- exprToDotPat' f
  case pf of 
     (ConP co c ps) -> do pe <- exprToDotPat' e'
                          return $ ConP co c (ps ++ [pe])
     _ -> tell (All False) >> return (DotP e)
exprToDotPat' e = tell (All False) >> return (DotP e)
-- exprToDotPat' e@(App f []) = exprToDotPat' f
-- exprToDotPat' e@(App f es) = do
--   p <- exprToDotPat' f
--   case p of 
--      (ConP co c ps) -> do ps' <- mapM exprToDotPat' es
--                           return $ ConP co c (ps ++ ps')
--      _ -> tell (All False) >> return (DotP e)

patternToExpr :: Pattern -> Expr
patternToExpr (VarP n)       = Var n
patternToExpr (SizeP m n)    = Var n
patternToExpr (ConP co n ps) = foldl App (con co n) (map patternToExpr ps)
-- patternToExpr (ConP co n ps) = Con co n `App` (map patternToExpr ps)
patternToExpr (SuccP p)      = Succ (patternToExpr p)
patternToExpr (UnusableP p)  = patternToExpr p
patternToExpr (DotP e)       = e
patternToExpr (ErasedP p)    = Erased $ patternToExpr p
patternToExpr (AbsurdP)      = Irr

-- admissible pattern ------------------------------------------------

-- completeP is used in admPattern, should not be True for UnusableP
completeP :: Pattern -> Bool
completeP (DotP _) = True
completeP (VarP _) = True
completeP SizeP{}  = True
completeP (UnusableP p) = completeP p
completeP (ErasedP p)   = completeP p
completeP _ = False

isDotPattern :: Pattern -> Bool
isDotPattern (DotP _ ) = True
isDotPattern _ = False

-- isSuccessorPattern is used in admPattern, should not be True for UnusableP
isSuccessorPattern :: Pattern -> Bool
isSuccessorPattern (SuccP _) = True
isSuccessorPattern (DotP (Succ _)) = True
isSuccessorPattern (ErasedP (SuccP _)) = True
isSuccessorPattern _ = False

shallowSuccP :: Pattern -> Bool
shallowSuccP (SuccP p) = isVarP p
   where isVarP :: Pattern -> Bool
         isVarP (VarP _) = True
         isVarP (DotP (Var _)) = True
         isVarP (ErasedP (VarP _)) = True
         isVarP _ = False
shallowSuccP (DotP (Succ (Var _))) = True
shallowSuccP _ = False

-- telescopes --------------------------------------------------------

teleToType' :: (Dec -> Dec) -> Telescope -> Type -> Type
teleToType' mod [] t = t
teleToType' mod ((dec,n,t):tel) t2 = Pi (mod dec) n t (teleToType' mod tel t2)

teleToType :: Telescope -> Type -> Type
teleToType = teleToType' id

teleToTypeErase :: Telescope -> Type -> Type
teleToTypeErase = teleToType' (\ dec -> dec { erased = True })

splitTeleType :: Int -> (Telescope,Type) -> (Telescope,Type)
splitTeleType 0 (tel,t) = (tel,t)
splitTeleType k (tel,(Pi dec n t t2)) = splitTeleType (k - 1) (tel ++ [(dec,n,t)],t2) 

-- | typeToTele ((x : A) -> (y : B) -> C) = ([(x,A),(y,B)], C)
typeToTele :: Type -> (Telescope, Type)
typeToTele t = ttt t []
    where 
      ttt :: Type -> Telescope -> (Telescope,Type)
      ttt (Pi dec n t' t2) tel = ttt t2 ((dec,n,t') : tel)
      ttt x tel = (reverse tel, x)                          
